@@ -1,34 +1,36 @@
 /*
  * ndisc.c - ICMPv6 neighbour discovery command line tool
- * $Id: ndisc.c 355 2006-10-07 15:02:56Z remi $
+ * $Id: ndisc.c 483 2007-08-08 15:09:36Z remi $
  */
 
-/***********************************************************************
- *  Copyright (C) 2004-2006 Rémi Denis-Courmont.                       *
- *  This program is free software; you can redistribute and/or modify  *
- *  it under the terms of the GNU General Public License as published  *
- *  by the Free Software Foundation; version 2 of the license.         *
- *                                                                     *
- *  This program is distributed in the hope that it will be useful,    *
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of     *
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.               *
- *  See the GNU General Public License for more details.               *
- *                                                                     *
- *  You should have received a copy of the GNU General Public License  *
- *  along with this program; if not, you can get it from:              *
- *  http://www.gnu.org/copyleft/gpl.html                               *
- ***********************************************************************/
+/*************************************************************************
+ *  Copyright © 2004-2006 Rémi Denis-Courmont.                           *
+ *  This program is free software: you can redistribute and/or modify    *
+ *  it under the terms of the GNU General Public License as published by *
+ *  the Free Software Foundation, versions 2 or 3 of the license.        *
+ *                                                                       *
+ *  This program is distributed in the hope that it will be useful,      *
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of       *
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
+ *  GNU General Public License for more details.                         *
+ *                                                                       *
+ *  You should have received a copy of the GNU General Public License    *
+ *  along with this program. If not, see <http://www.gnu.org/licenses/>. *
+ *************************************************************************/
 
 #ifdef HAVE_CONFIG_H
 # include <config.h>
 #endif
 
+#include <gettext.h>
+
 #include <stdio.h>
 #include <string.h>
 #include <stdlib.h> /* div() */
 #include <inttypes.h>
 #include <limits.h> /* UINT_MAX */
 #include <locale.h>
+#include <stdbool.h>
 
 #include <errno.h> /* EMFILE */
 #include <sys/types.h>
@@ -61,21 +63,6 @@
 # define AI_IDN 0
 #endif
 
-#ifndef RDISC
-# define NAME "ndisc"
-# define ND_TYPE_ADVERT ND_NEIGHBOR_ADVERT
-# define TYPE_NAME "Neighbor"
-# define NDISC_DEFAULT (NDISC_VERBOSE1 | NDISC_SINGLE)
-# ifdef __linux__
-#  include <sys/ioctl.h>
-# endif
-#else
-# define NAME "rdisc"
-# define ND_TYPE_ADVERT ND_ROUTER_ADVERT
-# define TYPE_NAME "Router"
-# define NDISC_DEFAULT NDISC_VERBOSE1
-#endif
-
 
 enum ndisc_flags
 {
@@ -120,7 +107,7 @@ getipv6byname (const char *name, const char *ifname, int numeric,
 }
 
 
-inline int
+static inline int
 setmcasthoplimit (int fd, int value)
 {
 	return setsockopt (fd, IPPROTO_IPV6, IPV6_MULTICAST_HOPS,
@@ -144,6 +131,10 @@ printmacaddress (const uint8_t *ptr, size_t len)
 
 
 #ifndef RDISC
+# ifdef __linux__
+#  include <sys/ioctl.h>
+# endif
+
 static int
 getmacaddress (const char *ifname, uint8_t *addr)
 {
@@ -176,6 +167,15 @@ getmacaddress (const char *ifname, uint8_t *addr)
 # endif
 }
 
+
+static const uint8_t nd_type_advert = ND_NEIGHBOR_ADVERT;
+static const unsigned nd_delay_ms = 1000;
+static const unsigned ndisc_default = NDISC_VERBOSE1 | NDISC_SINGLE;
+static const char ndisc_usage[] = N_(
+	"Usage: %s [options] <IPv6 address> <interface>\n"
+	"Looks up an on-link IPv6 node link-layer address (Neighbor Discovery)\n");
+static const char ndisc_dataname[] = N_("link-layer address");
+
 typedef struct
 {
 	struct nd_neighbor_solicit hdr;
@@ -183,6 +183,7 @@ typedef struct
 	uint8_t hw_addr[6];
 } solicit_packet;
 
+
 static ssize_t
 buildsol (solicit_packet *ns, struct sockaddr_in6 *tgt, const char *ifname)
 {
@@ -194,8 +195,8 @@ buildsol (solicit_packet *ns, struct sockaddr_in6 *tgt, const char *ifname)
 	memcpy (&ns->hdr.nd_ns_target, &tgt->sin6_addr, 16);
 
 	/* determines actual multicast destination address */
-	memcpy (&tgt->sin6_addr.s6_addr, "\xff\x02\x00\x00\x00\x00\x00\x00"
-	                                 "\x00\x00\x00\x01\xff", 13);
+	memcpy (tgt->sin6_addr.s6_addr, "\xff\x02\x00\x00\x00\x00\x00\x00"
+	                                "\x00\x00\x00\x01\xff", 13);
 
 	/* gets our own interface's link-layer address (MAC) */
 	if (getmacaddress (ifname, ns->hw_addr))
@@ -209,7 +210,7 @@ buildsol (solicit_packet *ns, struct sockaddr_in6 *tgt, const char *ifname)
 
 static int
 parseadv (const uint8_t *buf, size_t len, const struct sockaddr_in6 *tgt,
-         int verbose)
+          bool verbose)
 {
 	const struct nd_neighbor_advert *na =
 		(const struct nd_neighbor_advert *)buf;
@@ -261,11 +262,22 @@ parseadv (const uint8_t *buf, size_t len, const struct sockaddr_in6 *tgt,
 	return -1;
 }
 #else
+static const uint8_t nd_type_advert = ND_ROUTER_ADVERT;
+static const unsigned nd_delay_ms = 4000;
+static const unsigned ndisc_default = NDISC_VERBOSE1;
+static const char ndisc_usage[] = N_(
+	"Usage: %s [options] [IPv6 address] <interface>\n"
+	"Solicits on-link IPv6 routers (Router Discovery)\n");
+static const char ndisc_dataname[] = N_("advertised prefixes");
+
 typedef struct nd_router_solicit solicit_packet;
 
 static ssize_t
-buildsol (solicit_packet *rs)
+buildsol (solicit_packet *rs, struct sockaddr_in6 *tgt, const char *ifname)
 {
+	(void)tgt;
+	(void)ifname;
+
 	/* builds ICMPv6 Router Solicitation packet */
 	rs->nd_rs_type = ND_ROUTER_SOLICIT;
 	rs->nd_rs_code = 0;
@@ -275,8 +287,21 @@ buildsol (solicit_packet *rs)
 }
 
 
+static void
+print32time (uint32_t v)
+{
+	v = ntohl (v);
+
+	if (v == 0xffffffff)
+		fputs (_("    infinite (0xffffffff)\n"), stdout);
+	else
+		printf (_("%12u (0x%08x) %s\n"),
+		        v, v, ngettext ("second", "seconds", v));
+}
+
+
 static int
-parseprefix (const struct nd_opt_prefix_info *pi, size_t optlen, int verbose)
+parseprefix (const struct nd_opt_prefix_info *pi, size_t optlen, bool verbose)
 {
 	char str[INET6_ADDRSTRLEN];
 
@@ -294,24 +319,10 @@ parseprefix (const struct nd_opt_prefix_info *pi, size_t optlen, int verbose)
 
 	if (verbose)
 	{
-		/* INET6_ADDRSTRLEN > 13 */
-		unsigned v;
-
 		fputs (_("  Valid time              : "), stdout);
-		v = ntohl (pi->nd_opt_pi_valid_time);
-		if (v == 0xffffffff)
-			fputs (_("    infinite (0xffffffff)\n"), stdout);
-		else
-			printf (_("%12u (0x%08x) %s\n"),
-			        v, v, ngettext ("second", "seconds", v));
-
+		print32time (pi->nd_opt_pi_valid_time);
 		fputs (_("  Pref. time              : "), stdout);
-		v = ntohl (pi->nd_opt_pi_preferred_time);
-		if (v == 0xffffffff)
-			fputs (_("    infinite (0xffffffff)\n"), stdout);
-		else
-			printf (_("%12u (0x%08x) %s\n"),
-			        v, v, ngettext ("second", "seconds", v));
+		print32time (pi->nd_opt_pi_preferred_time);
 	}
 	return 0;
 }
@@ -329,13 +340,72 @@ parsemtu (const struct nd_opt_mtu *m)
 }
 
 
+static const char *
+pref_i2n (unsigned val)
+{
+	static const char *values[] =
+		{ N_("medium"), N_("high"), N_("medium (invalid)"), N_("low") };
+	return gettext (values[(val >> 3) & 3]);
+}
+
+
 static int
-parseadv (const uint8_t *buf, size_t len, int verbose)
+parseroute (const uint8_t *opt)
+{
+	uint8_t optlen = opt[1], plen = opt[2];
+	if ((optlen > 3) || (plen > 128) || (optlen < ((plen + 127) >> 6)))
+		return -1;
+
+	char str[INET6_ADDRSTRLEN];
+	struct in6_addr dst = in6addr_any;
+	memcpy (dst.s6_addr, opt + 8, (optlen - 1) << 3);
+	if (inet_ntop (AF_INET6, &dst, str, sizeof (str)) == NULL)
+		return -1;
+
+	printf (_(" Route                    : %s/%u\n"), str, (unsigned)plen);
+	printf (_("  Route preference        :       %6s\n"), pref_i2n (opt[3]));
+	fputs (_("  Route lifetime          : "), stdout);
+	print32time (((const uint32_t *)opt)[1]);
+	return 0;
+}
+
+
+static int
+parserdnss (const uint8_t *opt)
+{
+	uint8_t optlen = opt[1];
+	if (((optlen & 1) == 0) || (optlen < 3))
+		return -1;
+
+	optlen /= 2;
+	for (unsigned i = 0; i < optlen; i++)
+	{
+		char str[INET6_ADDRSTRLEN];
+
+		if (inet_ntop (AF_INET6, opt + (16 * i + 8), str,
+		               sizeof (str)) == NULL)
+			return -1;
+
+		printf (_(" Recursive DNS server     : %s\n"), str);
+	}
+
+	fputs (ngettext ("  DNS server lifetime     : ",
+	                 "  DNS servers lifetime    : ", optlen), stdout);
+	print32time (((const uint32_t *)opt)[1]);
+	return 0;
+}
+
+
+static int
+parseadv (const uint8_t *buf, size_t len, const struct sockaddr_in6 *tgt,
+          bool verbose)
 {
 	const struct nd_router_advert *ra =
 		(const struct nd_router_advert *)buf;
 	const uint8_t *ptr;
-	
+
+	(void)tgt;
+
 	/* checks if the packet is a Router Advertisement */
 	if ((len < sizeof (struct nd_router_advert))
 	 || (ra->nd_ra_type != ND_ROUTER_ADVERT)
@@ -347,8 +417,8 @@ parseadv (const uint8_t *buf, size_t len, int verbose)
 		unsigned v;
 
 		/* Hop limit */
-		fputs (_("\n"
-		         "Hop limit                 :    "), stdout);
+		puts ("");
+		fputs (_("Hop limit                 :    "), stdout);
 		v = ra->nd_ra_curhoplimit;
 		if (v != 0)
 			printf (_("      %3u"), v);
@@ -356,11 +426,18 @@ parseadv (const uint8_t *buf, size_t len, int verbose)
 			fputs (_("undefined"), stdout);
 		printf (_(" (      0x%02x)\n"), v);
 
+		v = ra->nd_ra_flags_reserved;
+		printf (_("Stateful address conf.    :          %3s\n"),
+		        gettext ((v & ND_RA_FLAG_MANAGED) ? N_ ("Yes") : N_("No")));
+		printf (_("Stateful other conf.      :          %3s\n"),
+		        gettext ((v & ND_RA_FLAG_OTHER) ? N_ ("Yes") : N_("No")));
+		printf (_("Router preference         :       %6s\n"), pref_i2n (v));
+
 		/* Router lifetime */
 		fputs (_("Router lifetime           : "), stdout);
 		v = ntohs (ra->nd_ra_router_lifetime);
 		printf (_("%12u (0x%08x) %s\n"), v, v,
-		        ngettext ("millisecond", "milliseconds", v));
+		        ngettext ("second", "seconds", v));
 
 		/* ND Reachable time */
 		fputs (_("Reachable time            : "), stdout);
@@ -396,15 +473,13 @@ parseadv (const uint8_t *buf, size_t len, int verbose)
 
 		len -= optlen;
 
-		/* skips unrecognized option */
-		switch (ptr[0])
+		/* only prefix are shown if not verbose */
+		switch (ptr[0] * (verbose ? 1
+		                          : (ptr[0] == ND_OPT_PREFIX_INFORMATION)))
 		{
 			case ND_OPT_SOURCE_LINKADDR:
-				if (verbose)
-				{
-					fputs (" Source link-layer address: ", stdout);
-					printmacaddress (ptr + 2, optlen - 2);
-				}
+				fputs (_(" Source link-layer address: "), stdout);
+				printmacaddress (ptr + 2, optlen - 2);
 				break;
 
 			case ND_OPT_TARGET_LINKADDR:
@@ -419,19 +494,24 @@ parseadv (const uint8_t *buf, size_t len, int verbose)
 				break; /* ignore */
 
 			case ND_OPT_MTU:
-				if (verbose)
-					parsemtu ((const struct nd_opt_mtu *)ptr);
+				parsemtu ((const struct nd_opt_mtu *)ptr);
+				break;
+
+			case 24: // RFC4191
+				parseroute (ptr);
+				break;
+
+			case 25: // RFC Ed queued draft-jeong-dnsop-ipv6-dns-discovery-12
+				parserdnss (ptr);
 				break;
 		}
+		/* skips unrecognized option */
 
 		ptr += optlen;
 	}
 
 	return 0;
 }
-
-# define buildsol( a, b, c ) buildsol (a)
-# define parseadv( a, b, c, d ) parseadv (a, b, d)
 #endif
 
 
@@ -535,7 +615,7 @@ recvadv (int fd, const struct sockaddr_in6 *tgt, unsigned wait_ms,
 		 && (addr.sin6_scope_id != tgt->sin6_scope_id))
 			continue;
 
-		if (parseadv (buf, val, tgt, flags & NDISC_VERBOSE) == 0)
+		if (parseadv (buf, val, tgt, (flags & NDISC_VERBOSE) != 0) == 0)
 		{
 			if (flags & NDISC_VERBOSE)
 			{
@@ -580,7 +660,7 @@ ndisc (const char *name, const char *ifname, unsigned flags, unsigned retry,
 		struct icmp6_filter f;
 
 		ICMP6_FILTER_SETBLOCKALL (&f);
-		ICMP6_FILTER_SETPASS (ND_TYPE_ADVERT, &f);
+		ICMP6_FILTER_SETPASS (nd_type_advert, &f);
 		setsockopt (fd, IPPROTO_ICMPV6, ICMP6_FILTER, &f, sizeof (f));
 	}
 
@@ -665,15 +745,7 @@ quick_usage (const char *path)
 static int
 usage (const char *path)
 {
-	printf (
-#ifndef RDISC
-_("Usage: %s [options] <IPv6 address> <interface>\n"
-"Looks up an on-link IPv6 node link-layer address (Neighbor Discovery)\n")
-#else
-_("Usage: %s [options] [IPv6 address] <interface>\n"
-"Solicits on-link IPv6 routers (Router Discovery)\n")
-#endif
-		, path);
+	printf (gettext (ndisc_usage), path);
 
 	printf (_("\n"
 "  -1, --single   display first response and exit\n"
@@ -685,13 +757,7 @@ _("Usage: %s [options] [IPv6 address] <interface>\n"
 "  -V, --version  display program version and exit\n"
 "  -v, --verbose  verbose display (this is the default)\n"
 "  -w, --wait     how long to wait for a response [ms] (default: 1000)\n"
-	           "\n"),
-#ifndef RDISC
-	           _("link-layer address")
-#else
-	           _("advertised prefixes")
-#endif
-		);
+	           "\n"), gettext (ndisc_dataname));
 
 	return 0;
 }
@@ -701,8 +767,9 @@ static int
 version (void)
 {
 	printf (_(
-NAME"6: IPv6 "TYPE_NAME" Discovery userland tool %s ($Rev: 355 $)\n"
-" built %s on %s\n"), VERSION, __DATE__, PACKAGE_BUILD_HOSTNAME);
+"ndisc6: IPv6 Neighbor/Router Discovery userland tool %s (%s)\n"), VERSION, "$Rev: 483 $");
+	printf (_(" built %s on %s\n"), __DATE__, PACKAGE_BUILD_HOSTNAME);
+
 	printf (_("Configured with: %s\n"), PACKAGE_CONFIGURE_INVOCATION);
 	puts (_("Written by Remi Denis-Courmont\n"));
 
@@ -714,7 +781,6 @@ NAME"6: IPv6 "TYPE_NAME" Discovery userland tool %s ($Rev: 355 $)\n"
 }
 
 
-
 static const struct option opts[] = 
 {
 	{ "single",   no_argument,       NULL, '1' },
@@ -741,10 +807,12 @@ main (int argc, char *argv[])
 	if (setuid (getuid ()) || ((fd >= 0) && (fd <= 2)))
 		return 1;
 
-	setlocale (LC_CTYPE, "");
+	setlocale (LC_ALL, "");
+	bindtextdomain (PACKAGE, LOCALEDIR);
+	textdomain (PACKAGE);
 
 	int val;
-	unsigned retry = 3, flags = NDISC_DEFAULT, wait_ms = 1000;
+	unsigned retry = 3, flags = ndisc_default, wait_ms = nd_delay_ms;
 	const char *hostname, *ifname;
 
 	while ((val = getopt_long (argc, argv, "1hmnqr:Vvw:", opts, NULL)) != EOF)
