@@ -1,34 +1,36 @@
 /*
  * traceroute.c - TCP/IPv6 traceroute tool
- * $Id: traceroute.c 360 2006-10-08 17:19:23Z remi $
+ * $Id: traceroute.c 483 2007-08-08 15:09:36Z remi $
  */
 
-/***********************************************************************
- *  Copyright (C) 2005-2006 Rémi Denis-Courmont.                       *
- *  This program is free software; you can redistribute and/or modify  *
- *  it under the terms of the GNU General Public License as published  *
- *  by the Free Software Foundation; version 2 of the license.         *
- *                                                                     *
- *  This program is distributed in the hope that it will be useful,    *
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of     *
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.               *
- *  See the GNU General Public License for more details.               *
- *                                                                     *
- *  You should have received a copy of the GNU General Public License  *
- *  along with this program; if not, you can get it from:              *
- *  http://www.gnu.org/copyleft/gpl.html                               *
- ***********************************************************************/
+/*************************************************************************
+ *  Copyright © 2005-2006 Rémi Denis-Courmont.                           *
+ *  This program is free software: you can redistribute and/or modify    *
+ *  it under the terms of the GNU General Public License as published by *
+ *  the Free Software Foundation, versions 2 or 3 of the license.        *
+ *                                                                       *
+ *  This program is distributed in the hope that it will be useful,      *
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of       *
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
+ *  GNU General Public License for more details.                         *
+ *                                                                       *
+ *  You should have received a copy of the GNU General Public License    *
+ *  along with this program. If not, see <http://www.gnu.org/licenses/>. *
+ *************************************************************************/
 
 #ifdef HAVE_CONFIG_H
 # include <config.h>
 #endif
 
+#include <gettext.h>
+
 #include <stdio.h>
 #include <string.h>
 #include <stdlib.h> /* div() */
 #include <limits.h>
 #include <stdbool.h>
 #include <time.h> /* nanosleep() */
+#include <assert.h>
 
 #include <unistd.h>
 #include <poll.h>
@@ -71,6 +73,10 @@
 # define IPV6_RECVHOPLIMIT IPV6_HOPLIMIT
 #endif
 
+#ifndef ICMP6_DST_UNREACH_BEYONDSCOPE
+# define ICMP6_DST_UNREACH_BEYONDSCOPE 2
+#endif
+
 #ifndef SOL_IPV6
 # define SOL_IPV6 IPPROTO_IPV6
 #endif
@@ -102,10 +108,32 @@ static uint16_t getsourceport (void)
 }
 
 
-ssize_t send_payload (int fd, const void *payload, size_t length)
+ssize_t send_payload (int fd, const void *payload, size_t length, int hlim)
 {
-	ssize_t rc = send (fd, payload, length, 0);
+	char cbuf[CMSG_SPACE (sizeof (int))];
+	struct iovec iov =
+	{
+		.iov_base = (void *)payload,
+		.iov_len = length
+	};
+	struct msghdr hdr =
+	{
+		.msg_name = NULL,
+		.msg_namelen = 0,
+		.msg_iov = &iov,
+		.msg_iovlen = 1,
+		.msg_control = cbuf,
+		.msg_controllen = sizeof (cbuf)
+	};
 
+	struct cmsghdr *cmsg = CMSG_FIRSTHDR (&hdr);
+	cmsg->cmsg_level = IPPROTO_IPV6;
+	cmsg->cmsg_type = IPV6_HOPLIMIT;
+	cmsg->cmsg_len = CMSG_LEN (sizeof (hlim));
+
+	memcpy (CMSG_DATA (cmsg), &hlim, sizeof (hlim));
+
+	ssize_t rc = sendmsg (fd, &hdr, 0);
 	if (rc == (ssize_t)length)
 		return 0;
 
@@ -151,7 +179,6 @@ recv_payload (int fd, void *buf, size_t len,
 }
 
 
-
 static bool has_port (int protocol)
 {
 	switch (protocol)
@@ -166,107 +193,39 @@ static bool has_port (int protocol)
 }
 
 
-/* Performs reverse lookup; print hostname and address */
+/**
+ * Stores difference between two struct timespec in a third struct timespec.
+ * The destination CAN be one (or both) term of the difference,
+ * and both terms can be identical (though this is rather pointless).
+ */
 static void
-printname (const struct sockaddr *addr, size_t addrlen)
+tsdiff (struct timespec *res,
+        const struct timespec *from, const struct timespec *to)
 {
-	char buf[NI_MAXHOST];
-
-	if (getnameinfo (addr, addrlen, buf, sizeof (buf), NULL, 0, niflags))
-		return;
-	printf (" %s", buf);
-
-	if (getnameinfo (addr, addrlen, buf, sizeof (buf), NULL, 0,
-	                 NI_NUMERICHOST | niflags))
-		return;
-	printf (" (%s) ", buf);
-}
-
-
-static inline void
-printipv6 (const struct sockaddr_in6 *addr)
-{
-	printname ((const struct sockaddr *)addr, sizeof (*addr));
-}
-
-
-/* Prints delay between two dates */
-static void
-printdelay (const struct timespec *from, const struct timespec *to)
-{
-	div_t d = div ((to->tv_nsec - from->tv_nsec) / 1000, 1000);
-
-	/*
-	 * For some stupid reasons, div() returns a negative remainder when
-	 * the numerator is negative, instead of following the mathematician
-	 * convention that the remainder always be positive.
-	 */
-	if (d.rem < 0)
+	res->tv_sec = to->tv_sec - from->tv_sec;
+	if (to->tv_nsec < from->tv_nsec)
 	{
-		d.quot--;
-		d.rem += 1000;
-	}
-	d.quot += 1000 * (to->tv_sec - from->tv_sec);
-
-	printf (_(" %u.%03u ms "), (unsigned)(d.quot), (unsigned)d.rem);
-}
-
-
-static inline void print_hlim (int hlim)
-{
-	if (hlim != -1)
-		printf (_("(%d) "), hlim);
-}
-
-
-static void
-print_icmp_code (const struct icmp6_hdr *hdr)
-{
-	if (hdr->icmp6_type == ICMP6_DST_UNREACH)
-	{
-		/* No path to destination */
-		char c = '\0';
-
-		switch (hdr->icmp6_code)
-		{
-			case ICMP6_DST_UNREACH_NOROUTE:
-				c = 'N';
-				break;
-
-			case ICMP6_DST_UNREACH_ADMIN:
-				c = 'S';
-				break;
-
-			case ICMP6_DST_UNREACH_ADDR:
-				c = 'H';
-				break;
-
-			case ICMP6_DST_UNREACH_NOPORT:
-				break;
-		}
-
-		if (c)
-			printf ("!%c ", c);
+		res->tv_sec--;
+		res->tv_nsec = 1000000000 + to->tv_nsec - from->tv_nsec;
 	}
+	else
+		res->tv_nsec = to->tv_nsec - from->tv_nsec;
 }
 
 
 static ssize_t
 parse (trace_parser_t func, const void *data, size_t len,
-       unsigned hlim, unsigned retry, uint16_t port)
+       int *hlim, int *attempt, uint16_t port)
 {
-	unsigned rhlim, rretry;
+	if (func == NULL)
+		return -1;
 
-	ssize_t rc = func (data, len, &rhlim, &rretry, port);
+	unsigned dummy = -1;
+	ssize_t rc = func (data, len, hlim, &dummy, port);
+	*attempt = dummy;
 	if (rc < 0)
 		return rc;
 
-	if (rhlim != hlim)
-		return -1;
-
-	if ((rretry != (unsigned)(-1)) && (rretry != retry))
-		return -1;
-
 	return rc;
 }
 
@@ -353,222 +312,405 @@ out:
 }
 
 
-static int
-probe_ttl (int protofd, int icmpfd, const struct sockaddr_in6 *dst,
-           unsigned ttl, unsigned retries, unsigned timeout, unsigned delay,
-           size_t plen)
+#define TRACE_TIMEOUT     0
+#define TRACE_OK          1 // TTL exceeded, echo reply, port unreachable...
+#define TRACE_CLOSED      2
+#define TRACE_OPEN        3
+#define TRACE_NOROUTE 0x100 // !N network unreachable
+#define TRACE_ADMIN   0x101 // !A administratively prohibited
+#define TRACE_BSCOPE  0X102 // !S beyond scope of source address
+#define TRACE_NOHOST  0x103 // !H address unreachable
+//#define TRACE_TOOBIG  0x100 // !F-%u packet too big
+//#define TRACE_NOSUP   0x400 // header field error
+#define TRACE_NOPROTO 0x401 // !P unrecognized next header
+//#define TRACE_NOOPT 0x402 // unrecognized option
+
+
+typedef struct
 {
-	struct in6_addr hop; /* hop if known from previous probes */
-	unsigned n;
-	int found = 0;
-	int state = -1; /* type of response received so far (-1: none,
-		0: normal, 1: closed, 2: open) */
-	/* see also: found (0: not found, <0: unreachable, >0: reached) */
-	struct timespec delay_ts;
+	struct sockaddr_in6 addr;  // hop address
+	struct timespec     sent;  // request date
+	struct timespec     rcvd;  // reply date
+	int                 rhlim; // received hop limit
+	unsigned            result;// 0: no reply, 1: ok, 2: closed, 3: open
+} tracetest_t;
+
+
+static int
+icmp_recv (int fd, tracetest_t *res, int *attempt, int *hlim,
+           const struct sockaddr_in6 *dst)
+{
+	struct
 	{
-		div_t d = div (delay, 1000);
-		delay_ts.tv_sec = d.quot;
-		delay_ts.tv_nsec = d.rem * 1000000;
-	}
+		struct icmp6_hdr hdr;
+		struct ip6_hdr inhdr;
+		uint8_t buf[1192];
+	} pkt;
+	res->rhlim = -1;
 
-	memset (&hop, 0, sizeof (hop));
-	printf ("%2d ", ttl);
-	setsockopt (protofd, SOL_IPV6, IPV6_UNICAST_HOPS, &ttl, sizeof (ttl));
+	ssize_t len = recv_payload (fd, &pkt, sizeof (pkt), &res->addr,
+	                            &res->rhlim);
 
-	for (n = 0; n < retries; n++)
+	if (len < (ssize_t)(sizeof (pkt.hdr) + sizeof (pkt.inhdr)))
+		return 0; // too small
+
+	len -= sizeof (pkt.hdr) + sizeof (pkt.inhdr);
+
+#if 0
+	/* "Extended" ICMP detection */
+	const uint8_t *ext = NULL;
+	size_t extlen = 0;
+	switch (pkt.hdr.icmp6_type)
 	{
-		struct timespec sent, recvd;
-
-		mono_gettime (&sent);
-		if (type->send_probe (protofd, ttl, n, plen, dst->sin6_port))
-		{
-			perror (_("Cannot send packet"));
-			return -1;
-		}
-
-		for (;;)
-		{
-			struct pollfd ufds[2];
-
-			memset (ufds, 0, sizeof (ufds));
-			ufds[0].fd = protofd;
-			ufds[0].events = POLLIN;
-			ufds[1].fd = icmpfd;
-			ufds[1].events = POLLIN;
-
-			mono_gettime (&recvd);
-			int val = ((sent.tv_sec + timeout - recvd.tv_sec) * 1000)
-				+ (int)((sent.tv_nsec - recvd.tv_nsec) / 1000000);
-
-			val = poll (ufds, 2, val > 0 ? val : 0);
-			if (val < 0) /* interrupted by signal - well, not really */
-				return -1;
-
-			if (val == 0)
+		case ICMP6_DST_UNREACH:
+		case ICMP6_TIME_EXCEEDED:
+			if (pkt.hdr.icmp6_data8[0] != 0)
 			{
-				fputs (" *", stdout);
-				break;
+				short origlen = pkt.hdr.icmp6_data8[0] << 6;
+				if (origlen > len)
+					return 0; // malformatted extended ICMP
+
+				assert (origlen >= 40);
+				ext = pkt.buf - 40 + origlen;
+				extlen = len - origlen;
+				len = origlen;
+
+				if ((extlen < 1) || ((ext[0] >> 4) != 2))
+					break; // no ext / unknown version
+				if (extlen < 4)
+					return 0; // malformatted v2 extension
+
+				/* FIXME: compute checksum */
 			}
-
-			mono_gettime (&recvd);
-
-			/* Receive final packet when host reached */
-			if (ufds[0].revents)
-			{
-				uint8_t buf[1240];
-				int hlim = -1;
-				ssize_t len;
-
-				len = recv_payload (protofd, buf, sizeof (buf), NULL, &hlim);
-				if (len < 0)
-				{
-					switch (errno)
-					{
-#ifdef EPROTO
-						case EPROTO:
-							/* Parameter problem seemingly can't be read from
-							 * the ICMPv6 socket, regardless of the filter. */
-							break;
+	}
 #endif
 
-						case EAGAIN:
-						case ECONNREFUSED:
-							continue;
+	const void *buf = skip_exthdrs (&pkt.inhdr, (size_t *)&len);
 
-						default:
-							/* These are very bad errors (-> bugs) */
-							perror (_("Receive error"));
-							return -1;
-					}
+	if (memcmp (&pkt.inhdr.ip6_dst, &dst->sin6_addr, 16))
+		return 0; // wrong destination
 
-					if (state == -1)
-					{
-						printipv6 (dst);
-						state = 1;
-					}
-					printdelay (&sent, &recvd);
-					found = ttl;
-					break; // response received, stop poll()ing
-				}
+	if (pkt.inhdr.ip6_nxt != type->protocol)
+		return 0; // wrong protocol
 
-				if (type->parse_resp == NULL)
-					continue;
+	len = parse (type->parse_err, buf, len, hlim, attempt, dst->sin6_port);
+	if (len < 0)
+		return 0;
 
-				len = parse (type->parse_resp, buf, len, ttl, n,
-				             dst->sin6_port);
-				if (len >= 0)
-				{
-					/* Route determination complete! */
-					if (state == -1)
-						printipv6 (dst);
+	/* interesting ICMPv6 error */
+	bool final = true;
+	switch (pkt.hdr.icmp6_type)
+	{
+		case ICMP6_DST_UNREACH:
+			switch (pkt.hdr.icmp6_code)
+			{
+				case ICMP6_DST_UNREACH_NOROUTE:
+				case ICMP6_DST_UNREACH_ADMIN:
+				case ICMP6_DST_UNREACH_BEYONDSCOPE:
+				case ICMP6_DST_UNREACH_ADDR:
+					res->result = 0x100 | pkt.hdr.icmp6_code;
+					break;
+				case ICMP6_DST_UNREACH_NOPORT:
+					res->result = TRACE_OK;
+					break;
+				case 5: // ingress/egress policy failure
+				case 6: // reject route
+					res->result = TRACE_ADMIN;
+					break;
+			}
+			break;
 
-					if (len != state)
-					{
-						const char *msg = NULL;
+		case ICMP6_PARAM_PROB:
+			switch (pkt.hdr.icmp6_code)
+			{
+				case ICMP6_PARAMPROB_NEXTHEADER:
+					res->result = 0x400 | pkt.hdr.icmp6_code;
+			}
+			break;
 
-						switch (len)
-						{
-							case 1:
-								msg = N_("closed");
-								break;
+		case ICMP6_TIME_EXCEEDED:
+			if (pkt.hdr.icmp6_code == ICMP6_TIME_EXCEED_TRANSIT)
+			{
+				res->result = TRACE_OK;
+				final = false; // intermediary reponse
+				break;
+			}
+		default: // should not happen (ICMPv6 filter)
+			return 0;
+	}
 
-							case 2:
-								msg = N_("open");
-								break;
-						}
+#if 0
+	/* "Extended" ICMP handling */
+	if (extlen > 0)
+	{
+		ext += 4;
+		extlen -= 4;
+		while (extlen >= 4)
+		{
+			uint16_t objlen = (ext[0] << 8) | ext[1];
+			if ((objlen & 3) || (objlen < 4) // malformatted object
+			 || (objlen > extlen)) // incomplete object
+				break;
 
-						if (msg != NULL)
-							printf ("[%s] ", msg);
-
-						state = len;
-					}
-
-					printdelay (&sent, &recvd);
-					print_hlim (hlim);
-					found = ttl;
-					break; // response received, stop poll()ing
-				}
+			switch (ext[2])
+			{
 			}
 
-			/* Receive ICMP errors along the way */
-			if (ufds[1].revents)
+			ext += objlen;
+			extlen -= objlen;
+		}
+	}
+#endif
+
+	if (!final)
+		return 1; // intermediary response
+
+	// final response received
+	return memcmp (&res->addr.sin6_addr, &dst->sin6_addr, 16) ? 2 : 3;
+}
+
+
+static int
+proto_recv (int fd, tracetest_t *res, int *attempt, int *hlim,
+            const struct sockaddr_in6 *dst)
+{
+	res->rhlim = -1;
+
+	uint8_t buf[1240];
+	ssize_t len = recv_payload (fd, buf, sizeof (buf), NULL,
+	                            &res->rhlim);
+	if (len < 0)
+	{
+		switch (errno)
+		{
+			case EAGAIN:
+			case ECONNREFUSED:
+#ifdef EPROTO
+			case EPROTO:
+#endif
+				return 0;
+
+			default:
+				/* These are very bad errors (-> bugs) */
+				fprintf (stderr, _("Receive error: %s\n"), strerror (errno));
+				return 0;
+		}
+	}
+
+	len = parse (type->parse_resp, buf, len, hlim, attempt, dst->sin6_port);
+	if (len < 0)
+		return 0;
+
+	/* Route determination complete! */
+	memcpy (&res->addr, dst, sizeof (res->addr));
+	res->result = 1 + len;
+	return 1; // response received
+}
+
+
+static int
+probe (int protofd, int icmpfd, const struct sockaddr_in6 *dst,
+       const struct timespec *deadline,
+       tracetest_t *res, int *hlim, int *attempt)
+{
+	for (;;)
+	{
+		struct pollfd ufds[2];
+
+		memset (ufds, 0, sizeof (ufds));
+		ufds[0].fd = protofd;
+		ufds[0].events = POLLIN;
+		ufds[1].fd = icmpfd;
+		ufds[1].events = POLLIN;
+
+		struct timespec recvd;
+		mono_gettime (&recvd);
+		int val = ((deadline->tv_sec  - recvd.tv_sec ) * 1000)
+		   + (int)((deadline->tv_nsec - recvd.tv_nsec) / 1000000);
+
+		val = poll (ufds, 2, val > 0 ? val : 0);
+		mono_gettime (&recvd);
+
+		if (val < 0) /* interrupted by signal - well, not really */
+			continue;
+		if (val == 0)
+		{
+			*hlim = -1;
+			break;
+		}
+
+		/* Receive final packet when host reached */
+		if (ufds[0].revents)
+		{
+			if (proto_recv (protofd, res, attempt, hlim, dst) > 0)
 			{
-				struct
-				{
-					struct icmp6_hdr hdr;
-					struct ip6_hdr inhdr;
-					uint8_t buf[1192];
-				} pkt;
-				struct sockaddr_in6 peer;
-				int hlim = -1;
-
-				ssize_t len = recv_payload (icmpfd, &pkt, sizeof (pkt),
-				                            &peer, &hlim);
-
-				if (len < (ssize_t)(sizeof (pkt.hdr) + sizeof (pkt.inhdr)))
-					continue; // too small
-
-				len -= sizeof (pkt.hdr) + sizeof (pkt.inhdr);
-
-				const void *buf = skip_exthdrs (&pkt.inhdr, (size_t *)&len);
-
-				if (memcmp (&pkt.inhdr.ip6_dst, &dst->sin6_addr, 16))
-					continue; // wrong destination
-
-				if (pkt.inhdr.ip6_nxt != type->protocol)
-					continue; // wrong protocol
-
-				len = parse (type->parse_err, buf, len, ttl, n,
-				             dst->sin6_port);
-				if (len < 0)
-					continue;
-
-				/* genuine ICMPv6 error that concerns us */
-				switch (pkt.hdr.icmp6_type)
-				{
-					case ICMP6_DST_UNREACH:
-						switch (pkt.hdr.icmp6_code)
-						{
-							case ICMP6_DST_UNREACH_NOPORT:
-								found = ttl;
-								break;
-
-							default:
-								if (found == 0)
-									found = -ttl;
-						}
-						break;
-
-					case ICMP6_TIME_EXCEEDED:
-						if (pkt.hdr.icmp6_code == ICMP6_TIME_EXCEED_TRANSIT)
-							break;
-
-					default: // should not happen (ICMPv6 filter)
-						continue;
-				}
-
-				if ((state == -1) || memcmp (&hop, &peer.sin6_addr, 16))
-				{
-					memcpy (&hop, &peer.sin6_addr, 16);
-					printipv6 (&peer);
-					state = 0;
-				}
-
-				printdelay (&sent, &recvd);
-				print_hlim (hlim);
-				print_icmp_code (&pkt.hdr);
-				break; // response received, stop poll()ing
+				res->rcvd = recvd;
+				return 1;
 			}
 		}
 
-		if (delay)
-			mono_nanosleep (&delay_ts);
+		/* Receive ICMP errors along the way */
+		if (ufds[1].revents)
+		{
+			val = icmp_recv (icmpfd, res, attempt, hlim, dst);
+			if (val)
+				res->rcvd = recvd;
+
+			switch (val)
+			{
+				case 1:
+					return 0; // TTL exceeded
+				case 2:
+					return -1; // unreachable
+				case 3:
+					return 1; // reached
+			}
+		}
 	}
-	puts ("");
-	return found;
+	return 0;
 }
 
 
+/* Performs reverse lookup; print hostname and address */
+static void
+printname (const struct sockaddr *addr, size_t addrlen)
+{
+	char buf[NI_MAXHOST];
+	int flags = niflags;
+
+	int c = getnameinfo (addr, addrlen, buf, sizeof (buf), NULL, 0, flags);
+	if (c == 0)
+	{
+		printf (" %s ", buf);
+
+		if (flags & NI_NUMERICHOST)
+			return;
+	}
+	flags |= NI_NUMERICHOST;
+
+	if (getnameinfo (addr, addrlen, buf, sizeof (buf), NULL, 0, flags))
+		strcpy (buf, "???");
+
+	if (c != 0)
+		printf (" %s ", buf); // work around DNS resolution failure
+
+	printf ("(%s) ", buf);
+}
+
+
+static inline void
+printrtt (const struct timespec *rtt)
+{
+	div_t d = div (rtt->tv_nsec / 1000, 1000);
+
+	/*
+	 * For some stupid reasons, div() returns a negative remainder when
+	 * the numerator is negative, instead of following the mathematicians
+	 * convention that the remainder always be positive.
+	 */
+	if (d.rem < 0)
+	{
+		d.quot--;
+		d.rem += 1000;
+	}
+	d.quot += 1000 * rtt->tv_sec;
+
+	printf (_(" %u.%03u ms "), (unsigned)(d.quot), (unsigned)d.rem);
+}
+
+
+static void
+display (const tracetest_t *tab, unsigned min_ttl, unsigned max_ttl,
+         unsigned retries)
+{
+	for (unsigned ttl = min_ttl; ttl <= max_ttl; ttl++)
+	{
+		struct sockaddr_in6 hop = { .sin6_family = AF_UNSPEC };
+		const tracetest_t *line = tab + retries * (ttl - min_ttl);
+
+		printf ("%2d ", ttl);
+
+		for (unsigned col = 0; col < retries; col++)
+		{
+			const tracetest_t *test = line + col;
+			if (test->result == TRACE_TIMEOUT)
+			{
+				fputs (" *", stdout);
+				continue;
+			}
+
+			if ((col == 0) || memcmp (&hop, &test->addr, sizeof (hop)))
+			{
+				memcpy (&hop, &test->addr, sizeof (hop));
+				printname ((struct sockaddr *)&hop, sizeof (hop));
+			}
+
+			struct timespec rtt;
+			tsdiff (&rtt, &test->sent, &test->rcvd);
+			printrtt (&rtt);
+
+			if ((col == 0) || (test[-1].result != test->result))
+			{
+				const char *msg;
+
+				switch (test->result)
+				{
+					case TRACE_CLOSED:
+						msg = N_("[closed] ");
+						break;
+
+					case TRACE_OPEN:
+						msg = N_("[open] ");
+						break;
+
+					default:
+						msg = NULL;
+				}
+
+				if (msg != NULL)
+					fputs (gettext (msg), stdout);
+			}
+
+			if (test->rhlim != -1)
+				printf ("(%d) ", test->rhlim);
+
+			const char *msg2;
+			switch (test->result)
+			{
+				case TRACE_NOROUTE:
+					msg2 = "!N ";
+					break;
+
+				case TRACE_ADMIN:
+					msg2 = "!A ";
+					break;
+
+				case TRACE_BSCOPE:
+					msg2 = "!S ";
+					break;
+
+				case TRACE_NOHOST:
+					msg2 = "!H ";
+					break;
+
+				case TRACE_NOPROTO:
+					msg2 = "!P ";
+					break;
+
+				default:
+					msg2 = NULL;
+			}
+
+			if (msg2 != NULL)
+				fputs (msg2, stdout);
+		}
+
+		fputc ('\n', stdout);
+	}
+}
+
+
+
 static int
 getaddrinfo_err (const char *host, const char *serv,
                  const struct addrinfo *hints, struct addrinfo **res)
@@ -594,7 +736,7 @@ connect_proto (int fd, struct sockaddr_in6 *dst,
 	memset (&hints, 0, sizeof (hints));
 	hints.ai_family = AF_INET6;
 	hints.ai_socktype = type->gai_socktype;
-	hints.ai_flags = AI_IDN;
+	hints.ai_flags = AI_CANONNAME | AI_IDN;
 
 	if ((srchost != NULL) || (srcport != NULL))
 	{
@@ -632,8 +774,12 @@ connect_proto (int fd, struct sockaddr_in6 *dst,
 	}
 
 	char buf[INET6_ADDRSTRLEN];
-	fputs (_("traceroute to"), stdout);
-	printname (res->ai_addr, res->ai_addrlen);
+	if (inet_ntop (AF_INET6,
+	               &((const struct sockaddr_in6 *)res->ai_addr)->sin6_addr,
+	               buf, sizeof (buf)) == NULL)
+		strcpy (buf, "??");
+
+	printf (_("traceroute to %s (%s) "), res->ai_canonname, buf);
 
 	if ((getsockname (fd, (struct sockaddr *)dst,
 	                  &(socklen_t){ sizeof (*dst) }) == 0)
@@ -642,10 +788,8 @@ connect_proto (int fd, struct sockaddr_in6 *dst,
 
 	memcpy (dst, res->ai_addr, res->ai_addrlen);
 	if (has_port (type->protocol))
-	{
-		printf (_("port %u, "), ntohs (dst->sin6_port));
-		printf (_("from port %u, "), ntohs (sport));
-	}
+		printf (_("port %u, from port %u, "), ntohs (dst->sin6_port),
+		        ntohs (sport));
 
 	freeaddrinfo (res);
 	return 0;
@@ -715,6 +859,7 @@ static struct
 	{ IPPROTO_ICMPV6, -1, EPERM },
 	{ IPPROTO_ICMPV6, -1, EPERM },
 	{ IPPROTO_UDP,    -1, EPERM },
+	{ IPPROTO_UDPLITE,-1, EPERM },
 	{ IPPROTO_TCP,    -1, EPERM }
 };
 
@@ -736,6 +881,7 @@ static int prepare_sockets (void)
 
 static int get_socket (int protocol)
 {
+	errno = EPROTONOSUPPORT;
 	for (unsigned i = 0; i < sizeof (protofd) / sizeof (protofd[0]); i++)
 		if (protofd[i].protocol == protocol)
 		{
@@ -764,7 +910,7 @@ static int
 traceroute (const char *dsthost, const char *dstport,
             const char *srchost, const char *srcport,
             unsigned timeout, unsigned delay, unsigned retries,
-            size_t packet_len, unsigned min_ttl, unsigned max_ttl)
+            size_t packet_len, int min_ttl, int max_ttl)
 {
 	/* Creates ICMPv6 socket to collect error packets */
 	int icmpfd = get_socket (IPPROTO_ICMPV6);
@@ -785,7 +931,7 @@ traceroute (const char *dsthost, const char *dstport,
 
 	drop_sockets ();
 
-#ifndef IPV6_PKTINFO
+#ifdef IPV6_PKTINFO
 	/* Set outgoing interface */
 	if (*ifname)
 	{
@@ -817,6 +963,7 @@ traceroute (const char *dsthost, const char *dstport,
 		ICMP6_FILTER_SETBLOCKALL (&f);
 		ICMP6_FILTER_SETPASS (ICMP6_DST_UNREACH, &f);
 		ICMP6_FILTER_SETPASS (ICMP6_TIME_EXCEEDED, &f);
+		ICMP6_FILTER_SETPASS (ICMP6_PARAM_PROB, &f);
 		setsockopt (icmpfd, SOL_ICMPV6, ICMP6_FILTER, &f, sizeof (f));
 	}
 
@@ -829,6 +976,17 @@ traceroute (const char *dsthost, const char *dstport,
 		goto error;
 	}
 
+	/* Set ICMPv6 filter for echo replies */
+	if (type->protocol == IPPROTO_ICMPV6)
+	{
+		// NOTE: we assume any ICMP probes type uses ICMP echo
+		struct icmp6_filter f;
+
+		ICMP6_FILTER_SETBLOCKALL (&f);
+		ICMP6_FILTER_SETPASS (ICMP6_ECHO_REPLY, &f);
+		setsockopt (protofd, SOL_ICMPV6, ICMP6_FILTER, &f, sizeof (f));
+	}
+
 #ifdef IPV6_TCLASS
 	/* Defines traffic class */
 	setsockopt (protofd, SOL_IPV6, IPV6_TCLASS, &tclass, sizeof (tclass));
@@ -842,33 +1000,147 @@ traceroute (const char *dsthost, const char *dstport,
 	if (rt_segc > 0)
 		setsock_rth (protofd, IPV6_RTHDR_TYPE_0, rt_segv, rt_segc);
 
-	/* Set ICMPv6 filter for echo replies */
-	if (type->protocol == IPPROTO_ICMPV6)
-	{
-		// This is ok as long as only one “type” uses ICMPv6 as protocol
-		struct icmp6_filter f;
-
-		ICMP6_FILTER_SETBLOCKALL (&f);
-		ICMP6_FILTER_SETPASS (ICMP6_ECHO_REPLY, &f);
-		setsockopt (protofd, SOL_ICMPV6, ICMP6_FILTER, &f, sizeof (f));
-	}
-
 	/* Defines destination */
 	struct sockaddr_in6 dst;
 	memset (&dst, 0, sizeof (dst));
 	if (connect_proto (protofd, &dst, dsthost, dstport, srchost, srcport))
 		goto error;
 	printf (_("%u hops max, "), max_ttl);
+
+	/* Adjusts packets length */
+	size_t overhead = sizeof (struct ip6_hdr);
+	if (rt_segc > 0)
+		overhead += inet6_rth_space (IPV6_RTHDR_TYPE_0, rt_segc);
+	if (packet_len < overhead)
+		packet_len = overhead;
+
 	printf (_("%lu byte packets\n"), (unsigned long)packet_len);
+	packet_len -= overhead;
+
+	struct timespec delay_ts;
+	if (delay)
+	{
+		div_t d = div (delay, 1000);
+		delay_ts.tv_sec = d.quot;
+		delay_ts.tv_nsec = d.rem * 1000000;
+	}
 
 	/* Performs traceroute */
 	int val = 0;
-	for (unsigned ttl = min_ttl; ttl <= max_ttl; ttl++)
+	if (max_ttl >= min_ttl)
 	{
-		val = probe_ttl (protofd, icmpfd, &dst, ttl,
-		                 retries, timeout, delay, packet_len);
-		if (val)
-			break;
+		tracetest_t tab[(1 + max_ttl - min_ttl) * retries];
+		memset (tab, 0, sizeof (tab));
+
+		for (unsigned step = 1, progress = 0;
+		     step < (1 + max_ttl - min_ttl) + retries;
+		     step++)
+		{
+			unsigned pending = 0;
+
+			if (isatty (1))
+			{
+				unsigned total = retries * (max_ttl - min_ttl + 1);
+				printf (_(" %3u%% completed..."), 100 * progress / total);
+				fputc ('\r', stdout);
+			}
+
+			if (delay && (step > 1))
+				mono_nanosleep (&delay_ts);
+
+			/* Sends requests */
+			for (unsigned i = 0; i < retries; i++)
+			{
+				int attempt = (retries - 1) - i;
+				int hlim = min_ttl + step + i - retries;
+
+				if ((hlim > max_ttl) || (hlim < min_ttl))
+					continue;
+
+				tracetest_t *t = tab + (hlim - min_ttl) * retries + attempt;
+				assert (t >= tab);
+				assert (t < tab + (sizeof (tab) / sizeof (tab[0])));
+
+				if (type->send_probe (protofd, hlim, attempt, packet_len,
+				                      dst.sin6_port))
+				{
+					fprintf (stderr, _("Cannot send data: %s\n"),
+					         strerror (errno));
+					return -1;
+				}
+
+				mono_gettime (&t->sent);
+				pending++;
+			}
+
+			struct timespec deadline;
+			mono_gettime (&deadline);
+			deadline.tv_sec += timeout;
+
+			/* Receives replies */
+			while (pending > 0)
+			{
+				tracetest_t results;
+				int hlim = -1;
+				int attempt = -1;
+				int res = probe (protofd, icmpfd, &dst, &deadline,
+				                 &results, &hlim, &attempt);
+
+				if (hlim == -1) /* timeout! */
+				{
+					progress += pending;
+					break;
+				}
+
+				if ((hlim > max_ttl) || (hlim < min_ttl))
+					continue;
+
+				if (attempt == -1)
+					attempt = min_ttl + step - (hlim + 1);
+
+				if ((attempt < 0) || ((unsigned)attempt >= retries))
+					continue;
+
+				tracetest_t *t = tab + (hlim - min_ttl) * retries + attempt;
+				assert (t >= tab);
+				assert (t < tab + (sizeof (tab) / sizeof (tab[0])));
+
+				if (t->result == TRACE_TIMEOUT /* no result yet */)
+				{
+					struct timespec buf = t->sent;
+					memcpy (t, &results, sizeof (*t));
+					t->sent = buf;
+					pending--;
+
+					if (isatty (1))
+					{
+						unsigned total = retries * (max_ttl - min_ttl + 1);
+						printf (_(" %3u%% completed..."),
+						        100 * ++progress / total);
+						fputc ('\r', stdout);
+					}
+				}
+
+				if (res && (val <= 0))
+				{
+					val = res > 0 ? 1 : -1; // sign <-> reachability
+					max_ttl = hlim;
+				}
+			}
+
+			if (isatty (1))
+			{
+				// white spaces to erase "xxx% completed..."
+				fputs (_("                  "), stdout);
+				fputc ('\r', stdout);
+			}
+
+			if (step >= retries)
+			{
+				int hl = min_ttl + step - retries;
+				display (tab + retries *(hl - min_ttl), hl, hl, retries);
+			}
+		}
 	}
 
 	/* Cleans up */
@@ -895,8 +1167,8 @@ static int
 usage (const char *path)
 {
 	printf (_(
-"Usage: %s [options] <IPv6 hostname/address> [port number/packet length]\n"
-"Print IPv6 network route to a host\n"), path);
+"Usage: %s [options] <IPv6 hostname/address> [%s]\n"
+"Print IPv6 network route to a host\n"), path, _("packet length"));
 
 	puts (_("\n"
 "  -A  send TCP ACK probes\n"
@@ -908,18 +1180,17 @@ usage (const char *path)
 "  -I  use ICMPv6 Echo Request packets as probes\n"
 "  -i  force outgoing network interface\n"
 "  -l  display incoming packets hop limit\n"
-/*"  -l  set TCP probes byte size\n"*/
 "  -m  set the maximum hop limit (default: 30)\n"
 "  -N  perform reverse name lookups on the addresses of every hop\n"
 "  -n  don't perform reverse name lookup on addresses\n"
-"  -p  override source TCP port or base destination UDP port\n"
+"  -p  override destination port\n"
 "  -q  override the number of probes per hop (default: 3)\n"
 "  -r  do not route packets\n"
 "  -S  send TCP SYN probes\n"
 "  -s  specify the source IPv6 address of probe packets\n"
 "  -t  set traffic class of probe packets\n"
 "  -U  send UDP probes (default)\n"
-"  -V, --version  display program version and exit\n"
+"  -V  display program version and exit\n"
 /*"  -v, --verbose  display all kind of ICMPv6 errors\n"*/
 "  -w  override the timeout for response in seconds (default: 5)\n"
 "  -z  specify a time to wait (in ms) between each probes (default: 0)\n"
@@ -933,8 +1204,8 @@ static int
 version (void)
 {
 	printf (_(
-"traceroute6: TCP & UDP IPv6 traceroute tool %s ($Rev: 360 $)\n"
-" built %s on %s\n"), VERSION, __DATE__, PACKAGE_BUILD_HOSTNAME);
+"traceroute6: TCP & UDP IPv6 traceroute tool %s (%s)\n"), VERSION, "$Rev: 483 $");
+	printf (_(" built %s on %s\n"), __DATE__, PACKAGE_BUILD_HOSTNAME);
 	printf (_("Configured with: %s\n"), PACKAGE_CONFIGURE_INVOCATION);
 	puts (_("Written by Remi Denis-Courmont\n"));
 
@@ -965,7 +1236,7 @@ parse_plen (const char *str)
 {
 	char *end;
 	unsigned long u = strtoul (str, &end, 0);
-	if ((u > 65535) || *end)
+	if ((u > 65575) || *end)
 	{
 		fprintf (stderr, _("%s: invalid packet length\n"), str);
 		return (size_t)(-1);
@@ -1005,7 +1276,7 @@ static const struct option opts[] =
 };
 
 
-static const char optstr[] = "AdEf:g:hIi:lm:Nnp:q:rSs:t:UVw:xz:";
+static const char optstr[] = "AdEFf:g:hIi:Llm:Nnp:q:rSs:t:UVw:xz:" "P:";
 
 int
 main (int argc, char *argv[])
@@ -1013,10 +1284,12 @@ main (int argc, char *argv[])
 	if (prepare_sockets () || setuid (getuid ()))
 		return 1;
 
-	setlocale (LC_CTYPE, "");
+	setlocale (LC_ALL, "");
+	bindtextdomain (PACKAGE, LOCALEDIR);
+	textdomain (PACKAGE);
 
-	const char *dsthost, *srchost = NULL, *xxxport = NULL;
-	size_t plen = 16;
+	const char *dsthost, *srchost = NULL, *dstport = "33434", *srcport = NULL;
+	size_t plen = 60;
 	unsigned retries = 3, wait = 5, delay = 0, minhlim = 1, maxhlim = 30;
 	int val;
 
@@ -1067,6 +1340,13 @@ main (int argc, char *argv[])
 				ifname[IFNAMSIZ - 1] = '\0';
 				break;
 
+			/* We should really have a generic option and
+			 * use getprotobyname() instead. Semantics of -L
+			 * will likely change in future versions!! */
+			case 'L':
+				type = &udplite_type;
+				break;
+
 			case 'l':
 				show_hlim = true;
 				break;
@@ -1088,8 +1368,12 @@ main (int argc, char *argv[])
 				niflags |= NI_NUMERICHOST | NI_NUMERICSERV;
 				break;
 
+			case 'P':
+				srcport = optarg;
+				break;
+
 			case 'p':
-				xxxport = optarg;
+				dstport = optarg;
 				break;
 
 			case 'q':
@@ -1163,31 +1447,20 @@ main (int argc, char *argv[])
 	if (type == NULL)
 		type = &udp_type;
 
-	const char *srcport = NULL, *dstport = NULL;
+	if (optind >= argc)
+		return quick_usage (argv[0]);
 
-	if (type->protocol == IPPROTO_TCP)
-		srcport = xxxport;
-	else
-		dstport = xxxport;
+	dsthost = argv[optind++];
 
 	if (optind < argc)
 	{
-		dsthost = argv[optind++];
-
-		if (optind < argc)
-		{
-			if (type->protocol == IPPROTO_TCP)
-				dstport = argv[optind++];
-			else
-			if ((plen = parse_plen (argv[optind++])) == (size_t)(-1))
-				return 1;
-		}
+		plen = parse_plen (argv[optind++]);
+		if (plen == (size_t)(-1))
+			return 1;
 	}
-	else
-		return quick_usage (argv[0]);
 
-	if (dstport == NULL)
-		dstport = (type->protocol == IPPROTO_TCP) ? "80" : "33434";
+	if (optind < argc)
+		return quick_usage (argv[0]);
 
 	setvbuf (stdout, NULL, _IONBF, 0);
 	return -traceroute (dsthost, dstport, srchost, srcport, wait, delay,
